/********************************************************************
 * tile map management routines
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 ********************************************************************/

#include "background.h"
#include "game.h"
#include "debug.h"

#define SCREEN_TILE_WIDTH	32
#define	SCREEN_TILE_HEIGHT	32


BackgroundStruct gBkg0;

/***********************************************************
 * init the global background settings
 **********************************************************/
void Background_InitSystem (void)
{	// no need for system init yet
	return;
}


/*************************************************************************************************
 * initialize a background
 *
 * IN:		bkgNum - which background is being initialized (0..3)
 * 			tileBlock - character block number (0..3) - what block in VRAM the tiles for the background go
 * 			mapBlock - screen block number (0..31) - what block in VRAM the tile map will go
 * 			priority - priority setting (0..3)
 *          flags - settings flags (BKG_MOSAIC, BKG_COLOR_MODE, BKG_WRAP)
 * 			size - tile map size (0..3) - see below
 *
 * NOTES:	1) this routine does NOT turn on the display of the tile map in the display mode register
 * 			2) possible size settings:
 *				0 = 32x32 tiles, or 16x16 if rotating background
 *				1 = 64x32 tiles, or 32x32 if rotating background
 *				2 = 32x64 tiles, or 64x64 if rotating background
 *				3 = 64x64 tiles, or 128x128 if rotating background
 *			3) flags:
 * 				BKG_MOSAIC - if set, the background uses mosaic mode
 * 				BKG_COLOR_MODE - if set, the background uses a 256 color palette (default is 16 color palettes)
 * 				BKG_WRAP - if set, the background wraps at the edges (for rotating backgrounds)
 * 			4) this routine does not change or clear the scroll values in pBkg
 *************************************************************************************************/
BackgroundStruct *Background_New (u8 bkgNum, u8 tileBlock, u8 mapBlock, u8 priority, u16 flags, u8 size)
{
BackgroundStruct *pBkg;

	// the game only uses one background... point us at its global data
	pBkg = &gBkg0;

	// setup the pBkg structure
	pBkg->pTiles = VRAM_TILE_ADDRESS(tileBlock);
	pBkg->pMap = VRAM_TILEMAP_ADDRESS(mapBlock);

	pBkg->bkgNum = bkgNum & 0x03;	// only 4 backgrounds possible in the GBA... make sure here that we never have an invalid value

	pBkg->pos.x = pBkg->pos.y = 0;
	pBkg->delta.x = pBkg->delta.y = 0;
	pBkg->scrollX = pBkg->scrollY = 0;

	// the flags input uses the same flags as the hardware settings, so use it as a base to build on
	flags |= priority & BKG_PRIORITY_MASK;
	flags |= (tileBlock<<BKG_BASE_BLOCK_SHIFT) & BKG_BASE_BLOCK_MASK;
	flags |= (mapBlock<<BKG_SCREEN_BASE_BLOCK_SHIFT) & BKG_SCREEN_BASE_BLOCK_MASK;
	flags |= (size<<BKG_MAP_SIZE_SHIFT) & BKG_MAP_SIZE_MASK;

	// load the settings to the appropriate background setup register
	*(pREG_BKG_0_SET + bkgNum) = flags;

	return(pBkg);
}


