/********************************************************************
 * game collision
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 *******************************************************************/

#include "collide.h"
#include "background.h"
#include "game.h"
#include "dat.h"
#include "debug.h"

/*****************************************************
 * collide a point with the background terrain
 *
 * IN:		pPos = position to collide against
 *
 * RETURNS:	0 = no collision
 * 			non-zero = type of terrain hit
 ****************************************************/
u32 Collide_PointToTerrain (s24f8_vect2D *pPos)
{
s32 posX, posY;
s32 tilePos;
u16 tileNum;

	// convert the position to a tile position on screen
	posX = (s24f8_to_int(pPos->x))>>3;
	posY = (s24f8_to_int(pPos->y))>>3;

	if (   (posX < 0)
		|| (posX >= gpGameLevelMap->sizeX)
		|| (posY < 0)
		|| (posY >= gpGameLevelMap->sizeY))
	{	//out of bounds
		return(TERRAINTYPE_NONE);
	}

	// get the tile number at that location
	tilePos = (gpGameLevelMap->sizeX * posY) + posX;
	tileNum = (gpGameLevelMap->pMap[tilePos]) & TILE_IDX_MASK;

	// convert it to a tile type
	switch (tileNum)
	{
	case TERRAINTILE_LADDER:
		return(TERRAINTYPE_LADDER);

	case TERRAINTILE_LADDERBASE:
		return(TERRAINTYPE_LADDERBASE);

	default:
		if ((tileNum >= TERRAINTILE_SOLID_START) && (tileNum < (TERRAINTILE_SOLID_START+TERRAINTILE_NUM_SOLID)))
		{	// solid
			return(TERRAINTYPE_SOLID);
		}

		// no collision
		return(TERRAINTYPE_NONE);
	}

}


/**********************************************************************
 * collide a point with a box
 *
 * IN:		pPos = position of the point
 * 			pBoxPos = the position of the box
 * 			pBoxMin = min extents of the box
 * 			pBoxMax = max extents of the box
 *
 * RETURNS:	non-zero if a collision happened
 *********************************************************************/
u32 Collide_PointToBox (s24f8_vect2D *pPos, s24f8_vect2D *pBoxPos, s24f8_vect2D *pBoxMin, s24f8_vect2D *pBoxMax)
{
s24f8_vect2D posRel;

	// convert point to be relative to the box
	posRel.x = pBoxPos->x - pPos->x;
	posRel.y = pBoxPos->y - pPos->y;

	// is the point inside the box?
	if (   (posRel.x < pBoxMin->x) || (posRel.x > pBoxMax->x)
		|| (posRel.y < pBoxMin->y) || (posRel.y > pBoxMax->y))
	{	// point is outside the box
		return(0);
	}

	return(1);
}


/**********************************************************************
 * collide a box with a box
 *
 * IN:		pBox1Pos = the position of box 1
 * 			pBox1Min = min extents of box 1
 * 			pBox1Max = max extents of box 1
 * 			pBox2Pos = the position of box 2
 * 			pBox2Min = min extents of box 2
 * 			pBox2Max = max extents of box 2
 *
 * RETURNS:	non-zero if a collision happened
 *********************************************************************/
u32 Collide_BoxToBox (s24f8_vect2D *pBox1Pos, s24f8_vect2D *pBox1Min, s24f8_vect2D *pBox1Max,
					  s24f8_vect2D *pBox2Pos, s24f8_vect2D *pBox2Min, s24f8_vect2D *pBox2Max)
{
s24f8_vect2D box1_TL, box1_BR, box2_TL, box2_BR;

	// calc extents of the boxes
	box1_TL.x = pBox1Pos->x + pBox1Min->x;
	box1_TL.y = pBox1Pos->y + pBox1Min->y;
	box1_BR.x = pBox1Pos->x + pBox1Max->x;
	box1_BR.y = pBox1Pos->y + pBox1Max->y;
	box2_TL.x = pBox2Pos->x + pBox2Min->x;
	box2_TL.y = pBox2Pos->y + pBox2Min->y;
	box2_BR.x = pBox2Pos->x + pBox2Max->x;
	box2_BR.y = pBox2Pos->y + pBox2Max->y;

	if (   (box1_BR.x < box2_TL.x)	// box 1 is to the left of box 2
		|| (box1_BR.y < box2_TL.y)	// box 1 is above box 2
		|| (box1_TL.x > box2_BR.x)	// box 1 is to the right of box 2
		|| (box1_TL.y > box2_BR.y))	// box 1 is below box 2
	{	// boxes are seperated... no collide
		return(0);
	}

	// we collided
	return(1);
}


// game object collide box sizes
s24f8_vect2D dudeMin = {int_to_s24f8(-2), int_to_s24f8(-14)};
s24f8_vect2D dudeMax = {int_to_s24f8(2), int_to_s24f8(-2)};

s24f8_vect2D powerupMin = {int_to_s24f8(0), int_to_s24f8(0)};
s24f8_vect2D powerupMax = {int_to_s24f8(6), int_to_s24f8(6)};

s24f8_vect2D shotMin = {float_to_s24f8(-1.5f), float_to_s24f8(-1.5f)};
s24f8_vect2D shotMax = {float_to_s24f8(1.5f), float_to_s24f8(1.5)};

/*************************************************************************
 * collide a powerup to the player
 *
 * IN:		pPowerup = ptr to a powerup
 *
 * RETURNS:	non-zero if the powerup was picked up
 *************************************************************************/
u32 Collide_PowerupToPlayer(PowerupStruct *pPowerup)
{
	if (Collide_BoxToBox(&pPowerup->pos, &powerupMin, &powerupMax, &gpPlayerDude->pos, &dudeMin, &dudeMax))
	{	// the boxes collided... give the player his powerup
		Powerup_GiveToDude(gpPlayerDude, pPowerup);
		return(1);
	}

	return(0);
}


/*******************************************************************
 * collide a bullet with the enemy dudes
 *
 * IN:		the bullet's position
 *
 * RETURNS:	non-zero if the bullet collided with an enemy
 ******************************************************************/
u32 Collide_BulletToEnemy (s24f8_vect2D *pPos)
{
u32 curObjID;
DudeStruct *pCurDude;

	curObjID = activeObjects[OBJECT_DUDE];

	while (curObjID != OBJECT_NULL)
	{
		pCurDude = (DudeStruct *)&gpObjectBuffer[curObjID];

		if (pCurDude != gpPlayerDude)
		{
			if (Collide_BoxToBox(pPos, &shotMin, &shotMax, &pCurDude->pos, &dudeMin, &dudeMax))
			{
				pCurDude->behaviorState = DUDE_STATE_DYING;
				return(1);
			}
		}

		curObjID = pCurDude->next;
	}

	return(0);
}


/*******************************************************************
 * collide the player with the enemy dudes
 *
 * RETURNS:	non-zero if the player collided with an enemy
 ******************************************************************/
u32 Collide_PlayerToEnemy (void)
{
u32 curObjID;
DudeStruct *pCurDude;

	curObjID = activeObjects[OBJECT_DUDE];

	while (curObjID != OBJECT_NULL)
	{
		pCurDude = (DudeStruct *)&gpObjectBuffer[curObjID];

		if (pCurDude != gpPlayerDude)
		{
			if (Collide_BoxToBox(&gpPlayerDude->pos, &dudeMin, &dudeMax, &pCurDude->pos, &dudeMin, &dudeMax))
			{
				return(1);
			}
		}

		curObjID = pCurDude->next;
	}

	return(0);
}


