/********************************************************************
 * dude game object
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 *******************************************************************/

#include "dude.h"
#include "sprite.h"
#include "buttons.h"
#include "collide.h"
#include "game.h"
#include "bullet.h"
#include "dat.h"
#include "debug.h"


#define	DUDE_WALK_FRAMEDIST		int_to_s24f8(2)
#define	DUDE_CLIMB_FRAMEDIST	int_to_s24f8(2)

#define	DUDE_WALK_SPEED			float_to_s24f8(0.5f)	// walk speed (in pixels per tick)
#define	DUDE_CLIMB_SPEED		float_to_s24f8(0.5f)	// climb speed
#define	DUDE_CLIMB_LAT_SPEED	float_to_s24f8(0.12f)	// edge laterally while climbing speed
#define	DUDE_FALL_LAT_SPEED		float_to_s24f8(0.5f)	// lateral speed in mid-air
#define	DUDE_GRAVITY			8 //float_to_s24f8(0.05f)	// gravity
#define	DUDE_JUMP_ACCEL			float_to_s24f8(0.6f)	// jump acceleration

#define	DUDE_SPRITE_POS_OFFSET_X	(-3)
#define	DUDE_SPRITE_POS_OFFSET_Y	(-15)


/***********************************************************
 * quick sprite display with flip, using dude settings:
 * 		- 8x16 (single tile) sprite size
 * 		- 16 color sprite
 *
 * IN:	screenX,Y - screen position of the sprite (in pixels, 0..255)
 *      u16 tileNum - tile number of the sprite graphics (0..1023)
 * 		u16 palNum - palette to use (0..15)
 * 		u16 flipBits - flip bits (SPRITE_HFLIP, SPRITE_VFLIP)
 ***********************************************************/
#define Sprite_DrawDude(screenX,screenY,tileNum,palNum,flipBits)					\
			Sprite_LoadNextSlot((u16)(SPRITE_SHAPE_WIDE + screenY),					\
								(u16)(SPRITE_SIZE_8 + (flipBits) + (screenX)), 		\
								(u16)((tileNum) + (palNum<<SPRITE_PALETTE_SHIFT)))	\


/************************************************************
 * create a new Dude character
 *
 * IN:		pPos = position to launch at
 * 			dudeType = type of dude
 *
 * RETURNS:	ptr to the new Dude game object
 ***********************************************************/
DudeStruct *Dude_New (s24f8_vect2D *pPos, u32 dudeType)
{
DudeStruct *pDude;
u32 objID;

	// grab a new object
	objID = Object_New(OBJECT_DUDE);
	if (objID == OBJECT_ID_NULL)
	{	// error... no unused objects left
		return(NULL);
	}

	// get the pointer to it
	pDude = (DudeStruct *)Object_getPtrS(objID);

	// init it
	pDude->dudeType = dudeType;
	pDude->pos = *pPos;
	pDude->lastFramePos = *pPos;
	pDude->behaviorState = DUDE_STATE_FALL;
	pDude->flip = Random_Get32() & 0x1;
	pDude->aim = pDude->flip ? AIM_LEFT : AIM_RIGHT;
	pDude->animFrame = 0;
	pDude->gravAccel = 0;
	pDude->deathTimer = 0;

	if (pDude->dudeType == DUDE_PLAYER)
	{
		pDude->palette = DUDE_PALETTE;
		gPlayerAmmo = 5;	// init the player's ammo

	}
	else
	{
		pDude->palette = ENEMY_PALETTE;
	}

	return(pDude);
}


/****************************************************************
 * force a dude to stay within the screen boundries
 *
 * IN:		ptr to a position
 *
 * RETURNS:	non-zero if we hit a horizontal world limit
 ***************************************************************/
u32 dude_CheckInsideScreenEdges (s24f8_vect2D *pPos)
{
	if (pPos->x < int_to_s24f8(3))
	{
		return(1);
	}
	else if (pPos->x > int_to_s24f8(240-3))
	{
		return(1);
	}

	if (pPos->y < int_to_s24f8(16))
	{
		return(1);
	}
	else if (pPos->y > int_to_s24f8(159))
	{
		return(1);
	}

	return(0);
}


/****************************************************************
 * force a dude to stay within the screen boundries
 *
 * IN:		ptr to a dude
 *
 * RETURNS:	non-zero if we hit a horizontal world limit
 ***************************************************************/
u32 dude_KeepInsideScreenEdges (DudeStruct *pDude)
{
u32 ret = 0;

	if (pDude->pos.x < int_to_s24f8(3))
	{
		pDude->pos.x = int_to_s24f8(3);
		ret = 1;
	}
	else if (pDude->pos.x > int_to_s24f8(240-3))
	{
		pDude->pos.x = int_to_s24f8(240-3);
		ret = 1;
	}

	if (pDude->pos.y < int_to_s24f8(16))
	{
		pDude->pos.y = int_to_s24f8(16);
	}
	else if (pDude->pos.y > int_to_s24f8(159))
	{
		pDude->pos.y = int_to_s24f8(159);

		// hit bottom of screen!  kills a dude!
		pDude->behaviorState = DUDE_STATE_DYING;
	}

	return(ret);
}


/****************************************************************
 * update a computer controlled dude
 *
 * IN:		ptr to the Dude
 *
 * RETURNS:	button controls for the Dude
 ***************************************************************/
u16 dude_UpdateAI (DudeStruct *pDude)
{
s24f8_vect2D pos;
u16 ret;

	if (pDude->behaviorState == DUDE_STATE_WALK)
	{
		if (pDude->flip)
		{
			pos.x = pDude->pos.x - float_to_s24f8(3.0f);
		}
		else
		{
			pos.x = pDude->pos.x + float_to_s24f8(3.0f);
		}

		pos.y = pDude->pos.y;

		if (dude_CheckInsideScreenEdges(&pos) || (Collide_PointToTerrain(&pos) == TERRAINTYPE_NONE))
		{	// we are about to walk off the edge
			if (pDude->flip)
			{
				pDude->flip = 0;
			}
			else
			{
				pDude->flip = SPRITE_HFLIP;
			}
		}

		if (pDude->flip)
		{
			ret = BUTTON_LEFT;
		}
		else
		{
			ret = BUTTON_RIGHT;
		}

		return(ret);
	}

	return(0);
}


/****************************************************************
 * update a dude character
 ***************************************************************/
u32 Dude_Update (ObjectStruct *pObj)
{
DudeStruct *pDude = (DudeStruct *)pObj;
u32 tileNum;
u32 lastFrameFlip;
s24f8 frameDist;
u32 terrainType;
u16 buttons;

	lastFrameFlip = pDude->flip;

	if (pDude->dudeType == DUDE_PLAYER)
	{	// player gets his dude controls from the GBA buttons
		buttons = buttonsHeld;
	}
	else
	{	// run AI on the dude
		buttons = dude_UpdateAI(pDude);
	}

	// control the dude
	switch(pDude->behaviorState)
	{
	case DUDE_STATE_WALK:
		// walking
		if (buttons & BUTTON_RIGHT)
		{	// going right
			pDude->pos.x += DUDE_WALK_SPEED * frameTime;
			pDude->flip = 0;
		}
		else if (buttons & BUTTON_LEFT)
		{	// going left
			pDude->pos.x -= DUDE_WALK_SPEED * frameTime;
			pDude->flip = SPRITE_HFLIP;
		}

		if (buttons & BUTTON_A)
		{	// player tried to jump
			pDude->gravAccel = -DUDE_JUMP_ACCEL;
			pDude->behaviorState = DUDE_STATE_FALL;
			break;
		}

		if (buttons & BUTTON_B)
		{	// we started firing
			pDude->behaviorState = DUDE_STATE_FIRE;
			pDude->aim = 0;
			break;
		}

		// keep us on-screen
		dude_KeepInsideScreenEdges(pDude);

		// run terrain collision
		terrainType = Collide_PointToTerrain(&pDude->pos);
		switch (terrainType)
		{
		case TERRAINTYPE_SOLID:
			// still on solid... most common case when last frame was also on solid
			break;

		case TERRAINTYPE_LADDER:
			// found a ladder below us... we might want to transfer to it
			if (buttons & BUTTON_DOWN)
			{	// switch to the ladder
				pDude->behaviorState = DUDE_STATE_CLIMB;
			}
			// fall through... the player might want to climb up the ladder

		case TERRAINTYPE_LADDERBASE:
			// found the base of a ladder... the player can switch to it
			if (buttons & BUTTON_UP)
			{	// switch to the ladder
				pDude->behaviorState = DUDE_STATE_CLIMB;
			}
			break;

		default:
			// in air... start us falling
			pDude->behaviorState = DUDE_STATE_FALL;
			break;
		}
		break;

	case DUDE_STATE_FIRE:
		// climbing
		if (buttons & BUTTON_DOWN)
		{	// fire down
			if (buttons & BUTTON_RIGHT)
			{	// fire down right
				pDude->flip = 0;
				pDude->aim = AIM_DIAGDOWNRIGHT;
			}
			else if (buttons & BUTTON_LEFT)
			{	// fire down left
				pDude->flip = SPRITE_HFLIP;
				pDude->aim = AIM_DIAGDOWNLEFT;
			}
			else
			{	// fire down
				pDude->aim = AIM_DOWN;
			}
		}
		else if (buttons & BUTTON_UP)
		{	// fire up
			if (buttons & BUTTON_RIGHT)
			{	// fire up right
				pDude->flip = 0;
				pDude->aim = AIM_DIAGUPRIGHT;
			}
			else if (buttons & BUTTON_LEFT)
			{	// fire up left
				pDude->flip = SPRITE_HFLIP;
				pDude->aim = AIM_DIAGUPLEFT;
			}
			else
			{	// fire up
				pDude->aim = AIM_UP;
			}
		}
		else if (buttons & BUTTON_RIGHT)
		{	// fire right
			pDude->flip = 0;
			pDude->aim = AIM_RIGHT;
		}
		else if (buttons & BUTTON_LEFT)
		{	// fire left
			pDude->flip = SPRITE_HFLIP;
			pDude->aim = AIM_LEFT;
		}
		else
		{	// no buttons... aim with current flip
			if (pDude->flip)
			{
				pDude->aim = AIM_LEFT;
			}
			else
			{
				pDude->aim = AIM_RIGHT;
			}
		}

		if ((buttons & BUTTON_B) == 0)
		{	// player let go of the fire button... fire the shot and go back to walking
			if (gPlayerAmmo > 0)
			{
				s24f8_vect2D pos, speed;

				pos.x = pDude->pos.x;
				pos.y = pDude->pos.y - int_to_s24f8(8);

				switch (pDude->aim)
				{
				case AIM_UP:
					speed.x = float_to_s24f8(0.0f);
					speed.y = float_to_s24f8(-3.0f);
					break;
				case AIM_DOWN:
					speed.x = float_to_s24f8(0.0f);
					speed.y = float_to_s24f8(3.0f);
					break;
				case AIM_LEFT:
					speed.x = float_to_s24f8(-3.0f);
					speed.y = float_to_s24f8(0.0f);
					break;
				case AIM_RIGHT:
					speed.x = float_to_s24f8(3.0f);
					speed.y = float_to_s24f8(0.0f);
					break;
				case AIM_DIAGUPLEFT:
					speed.x = float_to_s24f8(-2.0f);
					speed.y = float_to_s24f8(-2.0f);
					break;
				case AIM_DIAGUPRIGHT:
					speed.x = float_to_s24f8(2.0f);
					speed.y = float_to_s24f8(-2.0f);
					break;
				case AIM_DIAGDOWNLEFT:
					speed.x = float_to_s24f8(-2.0f);
					speed.y = float_to_s24f8(2.0f);
					break;
				default:
				case AIM_DIAGDOWNRIGHT:
					speed.x = float_to_s24f8(2.0f);
					speed.y = float_to_s24f8(2.0f);
					break;
				}

				Bullet_New(&pos, &speed);

				// only player's fire right now!
				if (gPlayerAmmo > 0)
					gPlayerAmmo--;
			}

			pDude->behaviorState = DUDE_STATE_WALK;
		}

		break;

	case DUDE_STATE_CLIMB:
		// climbing
		if (buttons & BUTTON_DOWN)
		{	// go down
			pDude->pos.y += DUDE_CLIMB_SPEED * frameTime;
		}
		else if (buttons & BUTTON_UP)
		{	// go up
			pDude->pos.y -= DUDE_CLIMB_SPEED * frameTime;
		}

		if (buttons & BUTTON_RIGHT)
		{	// edge right
			pDude->pos.x += DUDE_CLIMB_LAT_SPEED * frameTime;
			pDude->flip = 0;
		}
		else if (buttons & BUTTON_LEFT)
		{	// edge left
			pDude->pos.x -= DUDE_CLIMB_LAT_SPEED * frameTime;
			pDude->flip = SPRITE_HFLIP;
		}

		if (buttons & BUTTON_B)
		{	// we started firing
			pDude->behaviorState = DUDE_STATE_FIRE;
			pDude->aim = 0;
			break;
		}

		if (buttons & BUTTON_A)
		{	// player tried to jump
			pDude->gravAccel = -DUDE_JUMP_ACCEL;
			pDude->behaviorState = DUDE_STATE_FALL;
			break;
		}

		// keep us on-screen
		dude_KeepInsideScreenEdges(pDude);

		// run terrain collision
		terrainType = Collide_PointToTerrain(&pDude->pos);
		switch (terrainType)
		{
		case TERRAINTYPE_LADDER:
			// most common case... still on the ladder
			break;

		case TERRAINTYPE_SOLID:
		case TERRAINTYPE_LADDERBASE:
			// hit solid ground... stop us
			pDude->pos.y = int_to_s24f8(s24f8_to_int(pDude->pos.y));
			pDude->behaviorState = DUDE_STATE_WALK;
			break;

		default:
			// in air... start us falling
			pDude->behaviorState = DUDE_STATE_FALL;
			break;
		}
		break;

	case DUDE_STATE_FALL:
		// falling
		if (buttons & BUTTON_RIGHT)
		{	// going right
			pDude->pos.x += DUDE_WALK_SPEED * frameTime;
			pDude->flip = 0;
		}
		else if (buttons & BUTTON_LEFT)
		{	// going left
			pDude->pos.x -= DUDE_WALK_SPEED * frameTime;
			pDude->flip = SPRITE_HFLIP;
		}

		// drop us from gravity
		pDude->pos.y += pDude->gravAccel;

		// keep us on-screen
		dude_KeepInsideScreenEdges(pDude);

		// increase accel from gravity
		pDude->gravAccel += DUDE_GRAVITY;

		if (pDude->gravAccel < 0)
		{	// acceleration is upwards... don't collide
			break;
		}

		// run terrain collision
		terrainType = Collide_PointToTerrain(&pDude->pos);
		switch (terrainType)
		{
		case TERRAINTYPE_NONE:
			// most common case... still in air
			break;

		case TERRAINTYPE_SOLID:
		case TERRAINTYPE_LADDERBASE:
			// hit solid ground... stop us
			pDude->pos.y = int_to_s24f8(s24f8_to_int(pDude->pos.y));
			pDude->behaviorState = DUDE_STATE_WALK;
			pDude->gravAccel = 0;

			break;

		case TERRAINTYPE_LADDER:
			// hit a ladder
			if (pDude->dudeType == DUDE_ENEMY)
			{	// for now, enemies don't climb
				pDude->behaviorState = DUDE_STATE_WALK;
			}
			else
			{
				pDude->behaviorState = DUDE_STATE_CLIMB;
			}

			pDude->gravAccel = 0;

			break;

		default:
			break;
		}
		break;

	default:
	// case DUDE_STATE_DYING
		if (pDude->deathTimer == 0)
		{	// we just entered the death state... set the time and palette
			pDude->palette = DEREZ_PALETTE;
			pDude->deathTimer = float_to_s24f8(60.0f);
			if (pDude->dudeType == DUDE_PLAYER)
			{   // longer death for player
				pDude->deathTimer *= 2;
			}
		}
		else
		{
			pDude->deathTimer -= frameTimeF;
			if (pDude->deathTimer <= 0)
			{	// time to go away... kill this dude
				if (pDude->dudeType == DUDE_PLAYER)
				{	// the player died... signal the system to restart
					gameplayNewState = GAMEPLAY_NORMAL;

					if (gPlayerLives <= 0)
					{	// we are out of lives... go back to 1st level
						gameplayNewState = GAMEPLAY_GAMEOVER;
					}
					else
					{
						gPlayerLives--;
					}
				}
				return(1);
			}
		}
		break;
	}

	// collide the player with the enemy
	if (pDude->dudeType == DUDE_PLAYER)
	{
		if (Collide_PlayerToEnemy())
		{
			pDude->behaviorState = DUDE_STATE_DYING;
		}
	}

	// did we flip directions this frame?
	if (lastFrameFlip != pDude->flip)
	{	// we flipped... last frame distance is reset
		pDude->lastFramePos.x = pDude->pos.x;
		pDude->lastFramePos.y = pDude->pos.y;
	}

	switch (pDude->behaviorState)
	{
	case DUDE_STATE_WALK:
		// how far are we from our last frame switch?
		frameDist = pDude->pos.x - pDude->lastFramePos.x;
		if (frameDist < 0)
		{
			frameDist = -frameDist;
		}

		if (frameDist > DUDE_WALK_FRAMEDIST)
		{	// we crossed the threshold distance... flip to next frame
			if (pDude->animFrame >= DUDE_WALK_NUM_FRAMES-1)
			{
				pDude->animFrame = 0;
			}
			else
			{
				pDude->animFrame++;
			}

			// last frame distance is reset
			pDude->lastFramePos.x = pDude->pos.x;
			pDude->lastFramePos.y = pDude->pos.y;
		}

		tileNum = DUDE_WALK_ANIM_START + pDude->animFrame;

		break;

	case DUDE_STATE_CLIMB:
		// how far are we from our last frame switch?
		frameDist = pDude->pos.y - pDude->lastFramePos.y;
		if (frameDist < 0)
		{
			frameDist = -frameDist;
		}

		if (frameDist > DUDE_CLIMB_FRAMEDIST)
		{	// we crossed the threshold distance... flip to next frame
			if (pDude->animFrame >= DUDE_CLIMB_NUM_FRAMES-1)
			{
				pDude->animFrame = 0;
			}
			else
			{
				pDude->animFrame++;
			}

			// last frame distance is reset
			pDude->lastFramePos.x = pDude->pos.x;
			pDude->lastFramePos.y = pDude->pos.y;
		}

		tileNum = DUDE_CLIMB_ANIM_START + pDude->animFrame;

		break;

	case DUDE_STATE_FIRE:
		switch (pDude->aim)
		{
		case AIM_UP:
		case AIM_DOWN:
			tileNum = DUDE_FIRE_AIM_VERT;
			break;

		case AIM_RIGHT:
		case AIM_LEFT:
			tileNum = DUDE_FIRE_AIM_HORIZ;
			break;

		case AIM_DIAGUPRIGHT:
		case AIM_DIAGUPLEFT:
			tileNum = DUDE_FIRE_AIM_DIAGUP;
			break;

		case AIM_DIAGDOWNRIGHT:
		case AIM_DIAGDOWNLEFT:
			tileNum = DUDE_FIRE_AIM_DIAGDOWN;
			break;

		default:
			tileNum = DUDE_FIRE_AIM_VERT;
			break;
		}
		break;

	default:
	//case DUDE_STATE_FALL:
	//case DUDE_STATE_DYING:
		tileNum = DUDE_JUMP_ANIM_START;

		break;
	}

	tileNum <<= 1;	// 2 tiles per anim frame;

	// draw the dude sprite
	Sprite_DrawDude(s24f8_to_int(pDude->pos.x) + DUDE_SPRITE_POS_OFFSET_X,
					s24f8_to_int(pDude->pos.y) + DUDE_SPRITE_POS_OFFSET_Y,
					tileNum, pDude->palette, pDude->flip);

	return(0);
}



