/***************************************************************
 * game loop and related game code
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 **************************************************************/

#include "game.h"
#include "gba_disp.h"
#include "gba_palette.h"
#include "object.h"
#include "interrupt.h"
#include "math.h"
#include "world.h"
#include "sprite.h"
#include "text.h"
#include "buttons.h"
#include "powerup.h"
#include "dat.h"
#include "eramHeap.h"
#include "text.h"
#include "debug.h"

// number of vblanks the previous frame took
u16 frameTime = 0;
u16 frameTimeAccum = 0;	// vblank counter for current frame
s24f8 frameTimeF = 0;

// current gameplay state, and new state to switch to
u16 gameplayState = 0;
u16 gameplayNewState = 0;

// game time
u8 clockTicks = 0;
u8 seconds = 0;
u8 minutes = 0;
u8 hours = 0;

// random numbr generator seed value
s16 randomSeed = 0;

// current game level
u32 gGameLevel = 0;
BackgroundFrameStruct *gpGameLevelMap = &room1_frame_00;

// globals for the player's info
DudeStruct *gpPlayerDude = NULL;	// ptr to the player's game object
u32 gPlayerAmmo = 0;				// amount of ammo the player has left
u32 gPlayerLives = 3;				// number of lives the player has left

// in-game text messages
u8 msg_pause[] = "-PAUSE-";
u8 msg_playerdead[] = "-GAME OVER-";
u8 msg_pressStart[] = "PRESS START TO PLAY";

// timer used by the logo
s16 logoTimer = 0;

/*************************************************************
 * init the game for displaying the logo screen
 ************************************************************/
void Game_Init_ShellLogo (void)
{
u32 *pLoad32, *pCur32;
u16 *pLoad16, *pCur16;
s32 i, j;

	// halt all interrupts
	Interrupt_Disable(0);

	// init the data heap
	EramHeap_InitSystem();

	// init the game to the main menu
	Object_InitSystem();
	Background_InitSystem();

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_4 | DISP_FORCE_BLANK;

	// load the palette to VRAM
	pLoad32 = (u32 *)VRAM_PALETTES_TOP;
	pCur32 = (u32 *)LogoOpusGames_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bitmap
	pLoad16 = (u16 *)VRAM_TOP;
	pCur16 = (u16 *)LogoOpusGames_data;
	for (i=0; i<160; i++)
	{
		for (j=0; j<120; j++)
		{
			*pLoad16++ = *pCur16++;
		}
	}
	// set the display for basic tile diplay (mode 0) and enable tile map 0
	REG_DISPLAY_MODE = DISP_MODE_4 | DISP_MAP_2;

	logoTimer = 180;	// set timer for 3 seconds

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);
}


/*************************************************************
 * update the shell logo display
 ************************************************************/
void Game_Update_ShellLogo (void)
{
	logoTimer--;
	if (logoTimer <= 0)
	{	// time to switch to game menus
		gameplayNewState = GAMEPLAY_SHELL_MENU;
	}
}


/*************************************************************
 * load VRAM for game menus
 ************************************************************/
void Game_Init_ShellMenu (void)
{
u32 *pLoad32, *pCur32;
u16 *pLoad16, *pCur16;
u32 i;

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_FORCE_BLANK;

	// set background color to black
	pLoad16 = (u16 *)VRAM_MAP_PALETTES_TOP;
	pLoad16[0] = GBAPalette_MakeColor(0, 0, 0);

	// load the sprite palettes
	pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
	for (i=0; i<(16 * 3); i++)	// dude pals (player, enemy, derez... pals 0 to 2)
	{
		*pLoad16++ = 0;
	}
	for (i=0; i<16; i++)	// bullet
	{
		*pLoad16++ = 0;
	}
	pCur16 = (u16 *)powerup_palette;	// icons and powerups pal 4
	for (i=0; i<16; i++)
	{
		*pLoad16++ = *pCur16++;
	}
	for (i=0; i<16; i++)		// font pal 5
	{
		*pLoad16++ = 0;
	}

	// set dude colors
	pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
	pLoad16[(DUDE_PALETTE * 16) + 1] = GBAPalette_MakeColor(8, 8, 31);
	pLoad16[(ENEMY_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 8, 8);
	pLoad16[(DEREZ_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 31, 0);
	// set bullet colors
	pLoad16[(BULLET_PALETTE * 16) + 2] = GBAPalette_MakeColor(31, 31, 0);
	// set font colors
	pLoad16[(FONT_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 31, 31);
	pLoad16[(FONT_PALETTE * 16) + 2] = GBAPalette_MakeColor(15, 15, 15);

	// load sprite tiles
	pLoad32 = (u32 *)VRAM_SPRITE_TILES;

	// load the dude character and bullet tiles to sprite VRAM
	pCur32 = (u32 *)Dude_Tiles;
	for (i=0; i<(DUDE_NUM_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bullet tiles
	pCur32 = (u32 *)Bullet_Tiles;
	for (i=0; i<(BULLET_NUM_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the font and icon tiles
	pCur32 = (u32 *)font_tiles;
	for (i=0; i<((NUM_FONT_TILES+NUM_ICON_TILES) * 8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// init the sprites
	Sprite_InitSystem();

	// set the display
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_SPRITE_MAPPING_1D | DISP_SPRITES;

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);
}


/*******************************************************
 * update the game menus
 ******************************************************/
void Game_Update_ShellMenu (void)
{
	// adjust seed value for the random number generator (this ensures more
	// randomness in the generator once we get into gameplay)
	randomSeed++;

	// get button input
	Buttons_Update();

	// reset the sprite allocation for this frame
	Sprite_TopOfFrame();

	TextSpr_DrawString(72,  8, FONT_PALETTE, "BLOD-RUNNER");
	TextSpr_DrawString(80, 24, FONT_PALETTE,  "STARRING");
	TextSpr_DrawString(96, 32, FONT_PALETTE,    "DUDE");

	TextSpr_DrawString( 8, 64, FONT_PALETTE, "A - JUMP");
	TextSpr_DrawString( 8, 72, FONT_PALETTE, "B - FIRE");

	Sprite_DrawQuick(8, 88, AMMO_TILENUM, POWERUP_PALETTE);
	TextSpr_DrawString(24, 88, FONT_PALETTE, "- AMMO");
	Sprite_DrawQuick(8, 96, BOX_TILENUM, POWERUP_PALETTE);
	TextSpr_DrawString(24, 96, FONT_PALETTE, "- TREASURE - GRAB TO WIN");

	TextSpr_DrawString(48, 144, FONT_PALETTE, msg_pressStart);

	if (buttonsDown & BUTTON_START)
	{	// go to gameplay!
		gameplayNewState = GAMEPLAY_NORMAL;

		// init the random number generator
		Random_ReSeed(randomSeed);
	}

	// cheat codes
	if ((buttonsHeld & (BUTTON_L + BUTTON_R)) == (BUTTON_L + BUTTON_R))
	{	// cheat code activator is on....
		if (buttonsDown & BUTTON_A)
		{	// change start level
			gGameLevel++;
			if (gGameLevel >= NUM_GAME_LEVELS)
			{
				gGameLevel = 0;
			}
		}

		if (buttonsDown & BUTTON_B)
		{	// change start level
			gPlayerLives++;
			if (gPlayerLives > 10)
			{
				gPlayerLives = 0;
			}
		}
	}
}


/*************************************************************
 * load VRAM for gameplay
 ************************************************************/
void Game_LoadVRAMforGameplay (void)
{
u32 *pLoad32, *pCur32;
u16 *pLoad16, *pCur16;
BackgroundStruct *pBkg;
u32 i;

	// set the display for permanent vblank, so we can load VRAM
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_FORCE_BLANK;

	// load the map palette to VRAM (load to both palette 0 and 1)
	pLoad32 = (u32 *)VRAM_MAP_PALETTES_TOP;
	pCur32 = (u32 *)room_palette;
	for (i=0; i<128; i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// set background color to black
	pLoad16 = (u16 *)VRAM_MAP_PALETTES_TOP;
	pLoad16[0] = GBAPalette_MakeColor(0, 0, 0);

	// load the sprite palettes
	pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
	for (i=0; i<(16 * 3); i++)	// dude pals (player, enemy, derez... pals 0 to 2)
	{
		*pLoad16++ = 0;
	}
	for (i=0; i<16; i++)	// bullet
	{
		*pLoad16++ = 0;
	}
	pCur16 = (u16 *)powerup_palette;	// icons and powerups pal 4
	for (i=0; i<16; i++)
	{
		*pLoad16++ = *pCur16++;
	}
	for (i=0; i<16; i++)		// font pal 5
	{
		*pLoad16++ = 0;
	}

	// set dude colors
	pLoad16 = (u16 *)VRAM_SPRITE_PALETTES_TOP;
	pLoad16[(DUDE_PALETTE * 16) + 1] = GBAPalette_MakeColor(8, 8, 31);		// player is blue
	pLoad16[(ENEMY_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 8, 8);		// enemies are red
	pLoad16[(DEREZ_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 31, 0);	// dying dudes are yellow

	// set bullet colors
	pLoad16[(BULLET_PALETTE * 16) + 2] = GBAPalette_MakeColor(31, 31, 0);

	// set font colors
	pLoad16[(FONT_PALETTE * 16) + 1] = GBAPalette_MakeColor(31, 31, 31);
	pLoad16[(FONT_PALETTE * 16) + 2] = GBAPalette_MakeColor(15, 15, 15);

	// load sprite tiles
	pLoad32 = (u32 *)VRAM_SPRITE_TILES;

	// load the dude character and bullet tiles to sprite VRAM
	pCur32 = (u32 *)Dude_Tiles;
	for (i=0; i<(DUDE_NUM_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bullet tiles
	pCur32 = (u32 *)Bullet_Tiles;
	for (i=0; i<(BULLET_NUM_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the font and icon tiles
	pCur32 = (u32 *)font_tiles;
	for (i=0; i<((NUM_FONT_TILES+NUM_ICON_TILES) * 8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// launch object for background 0 with tile block 0, tile map block 31, 32x32 tiles size
	pBkg = Background_New(0, 0, 31, 0, 0, 0);

	// load the game bkg tiles
	pLoad32 = (u32 *)pBkg->pTiles;
	pCur32 = (u32 *)tileset_start;
	for (i=0; i<(NUM_BKG_TILES*8); i++)
	{
		*pLoad32++ = *pCur32++;
	}

	// load the bkg map and launch the game assets
	Level_LaunchAndDraw();
}


/*************************************************************
 * init the game setup
 ************************************************************/
void Game_Init (void)
{
	// set the current game level
	gpGameLevelMap = GameLevels[gGameLevel];

	// init the data heap
	EramHeap_InitSystem();

	// init the object and background system
	Object_InitSystem();
	Background_InitSystem();

	// NOTE: object system must be inited first, since this also sets up the background objects
	Game_LoadVRAMforGameplay();

	// init the sprites
	Sprite_InitSystem();

	// init the game world
	World_Init(0, 0, int_to_s24f8(1024), int_to_s24f8(1024));

	// set the display for basic tile diplay (mode 0) and enable tile map 0
	REG_DISPLAY_MODE = DISP_MODE_0 | DISP_MAP_0 | DISP_SPRITE_MAPPING_1D | DISP_SPRITES;

	// set the tranparency of the backgrounds
	// I will need this when I set up two backgrounds
//	REG_BLEND_SETTINGS = BLEND_FIRST_BG0 | BLEND_SECOND_BG1 | BLEND_SECOND_OBJ | BLEND_ALPHA_NORMAL;
//	SetBlendLevel(0, 15);

	// start the interrupts (vblank only)
	Interrupt_Enable(INTERRUPT_VBLANK);

	frameTimeAccum = 0;

	clockTicks = 0;
	seconds = 0;
	minutes = 0;
	hours = 0;
}


/**************************************************************
 * the master game loop
 *
 * NOTE:	one pass through the loop = one game frame
 *************************************************************/
void Game_MainLoop (void)
{
	// set the starting state of the game
	gameplayNewState = GAMEPLAY_SHELL_LOGO;

	while (1)
	{
		if (gameplayNewState != GAMEPLAY_NONE)
		{	// we are switching to a new game mode
			switch (gameplayNewState)
			{	// init the new gameplay mode
			case GAMEPLAY_SHELL_LOGO:
				Game_Init_ShellLogo();
				break;

			case GAMEPLAY_SHELL_MENU:
				gGameLevel = 0;
				gPlayerLives = 3;

				Game_Init_ShellMenu();
				break;

			case GAMEPLAY_NORMAL:
				switch (gameplayState)
				{
				case GAMEPLAY_PAUSED:
					// we are going back to a currently running game... don't reset
					break;

				default:
					// load and start game
					Game_Init();
					break;
				}
				break;

			case GAMEPLAY_GAMEOVER:
			default:
				break;
			}

			gameplayState = gameplayNewState;
			gameplayNewState = GAMEPLAY_NONE;
		}

		// make sure at least one vblank occured during last frame
		while (frameTimeAccum == 0);

		// set the new frame counters
		frameTime = frameTimeAccum;
		frameTimeF = int_to_s24f8(frameTime);
		frameTimeAccum = 0;

		// reset the sprite allocation for this frame
		Sprite_TopOfFrame();

		switch (gameplayState)
		{
		case GAMEPLAY_NORMAL:
			// adjust game clock
			clockTicks += frameTime;
			if (clockTicks >= 60)
			{
				clockTicks -= 60;
				seconds++;

				if (seconds >= 60)
				{
					seconds -= 60;
					minutes++;

					if (minutes >= 60)
					{
						minutes -= 60;
						hours++;
					}
				}
			}

			Sprite_DrawQuick(4, 4, SKULL_TILENUM, POWERUP_PALETTE);
			TextSpr_DrawNumber(12, 4, FONT_PALETTE, 0, gPlayerLives);

			// draw ammo count on the hud
			Sprite_DrawQuick(28, 4, AMMO_TILENUM, POWERUP_PALETTE);
			TextSpr_DrawNumber(36, 4, FONT_PALETTE, 0, gPlayerAmmo);

			// get joypad input
			Buttons_Update();

			// update all the active game objects
			Object_UpdateAll();

			if (buttonsDown & BUTTON_START)
			{   // pause the game
				gameplayNewState = GAMEPLAY_PAUSED;
			}

			break;

		case GAMEPLAY_PAUSED:
			// just wait for unpause
			Buttons_Update();

			if (buttonsDown & BUTTON_START)
			{
				gameplayNewState = GAMEPLAY_NORMAL;
			}

			// put up the paused message
	 		TextSpr_DrawString(96, 76, FONT_PALETTE, msg_pause);

			break;

		case GAMEPLAY_SHELL_LOGO:
			Game_Update_ShellLogo();

			break;

		case GAMEPLAY_SHELL_MENU:
			Game_Update_ShellMenu();

			break;

		case GAMEPLAY_GAMEOVER:
			// display win screen
			Buttons_Update();

			TextSpr_DrawString(80, 76, FONT_PALETTE, msg_playerdead);
			TextSpr_DrawString(48, 86, FONT_PALETTE, msg_pressStart);

			if (buttonsDown & BUTTON_START)
			{	// go back to the shell screens
				gameplayNewState = GAMEPLAY_SHELL_LOGO;
			}

			// keep the enemies running around
			Object_UpdateAll();

			break;

		default:
			break;
		}

		// tell system to load the current sprite frame next vblank
		Sprite_EndOfFrame();
	}
}


