/************************************************************
 * headers for sprite handling
 *
 * written by Doug Lanford, opus@opusgames.com
 * for more info, go to www.opusgames.com
 *
 * This file is provided as freeware
 ***********************************************************/

#ifndef SPRITE_H
#define SPRITE_H

#include "types.h"
#include "math.h"
#include "gba_sprite.h"

// RAM copy of the sprite OAM table (copies to VRAM each frame)
extern SpriteOAMStruct spriteOAMtable[2][NUM_OAM_SPRITES];

extern u8 allocSpriteNum[2];
extern u8 spriteCurBuffer;
extern u8 loadOAMflag;


/***********************************************************
 * quick sprite display, using default settings:
 * 		- 8x8 (single tile) sprite size
 * 		- 16 color sprite
 * 		- no flip
 *
 * IN:	u16 screenX,Y - screen position of the sprite (in pixels)
 *      u16 tileNum - tile number of the sprite graphics
 * 		u16 palNum - palette to use (0..15)
 ***********************************************************/
#define Sprite_DrawQuick(screenX,screenY,tileNum,palNum)							\
			Sprite_LoadNextSlot((u16)(screenY),										\
								(u16)(SPRITE_SIZE_8 + (screenX)), 					\
								(u16)((tileNum) + (palNum<<SPRITE_PALETTE_SHIFT)))	\

/***********************************************************
 * quick sprite display with flip, using default settings:
 * 		- 8x8 (single tile) sprite size
 * 		- 16 color sprite
 *
 * IN:	screenX,Y - screen position of the sprite (in pixels, 0..255)
 *      u16 tileNum - tile number of the sprite graphics (0..1023)
 * 		u16 palNum - palette to use (0..15)
 * 		u16 flipBits - flip bits (SPRITE_HFLIP, SPRITE_VFLIP)
 ***********************************************************/
#define Sprite_DrawQuickFlip(screenX,screenY,tileNum,palNum,flipBits)				\
			Sprite_LoadNextSlot((u16)(screenY), 									\
								(u16)(SPRITE_SIZE_8 + (flipBits) + (screenX)), 		\
								(u16)((tileNum) + (palNum<<SPRITE_PALETTE_SHIFT)))	\



/***********************************************************
 * init the sprite system
 **********************************************************/
extern void Sprite_InitSystem (void);


/***********************************************************
 * get the next hardware sprite slot, and load it with the
 * input settings
 *
 * IN:		attributes0,1,2 of the sprite
 *
 * RETURNS:	number of the sprite slot used (0xFFFF is error)
 **********************************************************/
extern u16 Sprite_LoadNextSlot (u16 attribute0, u16 attribute1, u16 attribute2);


/********************************************************************
 * check if a sprite is on-screen
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *
 * RETURNS:	non-zero if the sprite is on-screen
 *******************************************************************/
extern u32 Sprite_CheckOnScreen (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos);


/********************************************************************
 * update a sprite in the sprite OAM table, calculating its screen
 * position from world position
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *			attribute0, attribute1, attribute2 - basic attribute data (without sprite position)
 *
 * NOTE:	assumes the sprite is on-screen!
 *******************************************************************/
extern void Sprite_GetPosAndUpdate (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos,
									u16 attribute0, u16 attribute1, u16 attribute2);


/********************************************************************
 * check if a sprite is on-screen, and if so then update it in the
 * sprite OAM table
 *
 * IN:		pPos = ptr to the position of the sprite (in s24f8_vect2D)
 * 			pSize = ptr to the sprite size (in s24f8_vect2D)
 * 			pWorldPos = ptr to the screen position (in s24f8_vect2D)
 *			attribute0, attribute1, attribute2 - basic attribute data (without sprite position)
 * RETURNS:	non-zero if the sprite is on-screen
 *******************************************************************/
extern u32 Sprite_OnScreenUpdate (s24f8_vect2D *pPos, s24f8_vect2D *pSize, s24f8_vect2D *pWorldPos,
								  u16 attribute0, u16 attribute1, u16 attribute2);


/***********************************************************
 * sprite code called at top of each frame
 **********************************************************/
extern void Sprite_TopOfFrame (void);


/***********************************************************
 * sprite code called at end of each frame
 **********************************************************/
extern void Sprite_EndOfFrame (void);


#endif

